from optparse import OptionParser

from parsley.config import *
from parsley.loggingex import *

from util import *

log = logging.getLogger("cumin.config")

class CuminConfig(Config):
    def __init__(self):
        super(CuminConfig, self).__init__()

        hdef = os.path.normpath("/usr/share/cumin")
        self.home = os.environ.get("CUMIN_HOME", hdef)

        if not os.path.isdir(self.home):
            raise Exception("Home path '%s' is not a directory")

        test = CuminConfigSection(self, "test")
        test.log_file.default = os.path.join(self.home, "log", "test.log")

    def create_web_section(self, name, strict_section):
        web = CuminConfigSection(self, name, strict_section)
        web.log_file.default = os.path.join(self.home, "log", name + ".log")

        param = ConfigParameter(web, "update-interval", int)
        param.default = 10

        param = ConfigParameter(web, "max-qmf-table-sort", int)
        param.default = 1000

        param = ConfigParameter(web, "host", str)
        param.default = "localhost"

        param = ConfigParameter(web, "port", int)
        param.default = 45672

        param = ConfigParameter(web, "operator-email", str)

        param = ConfigParameter(web, "user", str)

        param = ConfigParameter(web, "request-memory",    int)
        param.default = 512 # MB
        param = ConfigParameter(web, "request-memory-vm", int)
        param.default = 512 # MB
        param = ConfigParameter(web, "request-disk",      int)
        param.default = 1024 #MB
        param = ConfigParameter(web, "request-disk-vm",   int)
        param.default = 5 * 1024 #MB

        param = ConfigParameter(web, "persona", str)
        param.default = "default"
        
        param = ConfigParameter(web, "fast-view-attributes", str)
        param.default = "JobStatus,Cmd,Args,ExitStatus,JobStartDate,LastRemoteHost,LastJobStatus,Owner"

    def create_data_section(self, name, strict_section):
        data = CuminConfigSection(self, name, strict_section)
        data.log_file.default = os.path.join(self.home, "log", name + ".log")
        
        param = ConfigParameter(data, "include-classes", str)
        param = ConfigParameter(data, "exclude-classes", str)
        param = ConfigParameter(data, "agents", str)

        param = ConfigParameter(data, "expire-enabled", bool)
        param.default = True

        param = ConfigParameter(data, "expire-interval", int)
        param.default = 60 * 60 # 1 hour

        param = ConfigParameter(data, "expire-threshold", int)
        param.default = 24 * 60 * 60 # 1 day

        param = ConfigParameter(data, "vacuum-enabled", bool)
        param.default = True

        param = ConfigParameter(data, "vacuum-interval", int)
        param.default = 60 * 60 # 1 hour

    def create_master_section(self, name, strict_section):
        '''
        Creates a blank master section.
        CuminConfigSection values are not interesting to the master
        '''

        master = ConfigSection(self, name, strict_section)
        param = ConfigParameter(master, "datas", str)
        param.default = "data"

        param = ConfigParameter(master, "webs", str)
        param.default = "web"
        
    def get_home(self):
        return self.home

    def parse(self):
        paths = list()

        paths.append(os.path.join(os.sep, "etc", "cumin", "cumin.conf"))
        paths.append(os.path.join(self.home, "etc", "cumin.conf"))
        paths.append(os.path.join(os.path.expanduser("~"), ".cumin.conf"))

        return self.parse_files(paths)

class CuminCommonConfig(CuminConfig):
    '''
    Provide a simple class which just gives access
    to the values in the common section of the config
    file for apps which do not need any more
    '''
    def __init__(self):
        super(CuminCommonConfig, self).__init__()

        self.section = CuminConfigSection(self, "common")

    def set_log(self, log):
        self.section.log_file.default = log

class CuminWebConfig(CuminConfig):
    def __init__(self, section_name="web", strict_section="False"):
        super(CuminWebConfig, self).__init__()

        self.create_web_section(section_name, strict_section)
    
class CuminDataConfig(CuminConfig):
    def __init__(self, section_name="data", strict_section=False):
        super(CuminDataConfig, self).__init__()

        self.create_data_section(section_name, strict_section)

class CuminMasterConfig(CuminConfig):
    def __init__(self, section_name="master", strict_section=False):
        super(CuminMasterConfig, self).__init__()

        self.create_master_section(section_name, strict_section)

class CuminConfigSection(ConfigSection):
    def __init__(self, config, name, strict_section=False):
        super(CuminConfigSection, self).__init__(config, name, strict_section)

        param = ConfigParameter(self, "database", str)
        param.default = "dbname=cumin user=cumin host=localhost"

        param = ConfigParameter(self, "brokers", str)
        param.default = "amqp://localhost"

        # Leave default set to None, which is equivalent to 
        # previous behavior
        param = ConfigParameter(self, "sasl-mech-list", str)

        self.log_file = ConfigParameter(self, "log-file", str)

        param = ConfigParameter(self, "log-level", str)
        param.default = "info"

        param = ConfigParameter(self, "log-max-mb", float)
        param.default = 10

        param = ConfigParameter(self, "log-max-archives", int)
        param.default = 1

        param = ConfigParameter(self, "debug", bool)
        param.default = False

class CuminOptionParser(OptionParser,object):
    def __init__(self):
        OptionParser.__init__(self)

        self.add_option("--init-only", action="store_true", default=False)

        # Defaults for these come later from the config
        # section named in --section, via apply_defaults()
        self.add_option("--debug", action="store_true")
        self.add_option("--database")
        self.add_option("--brokers")
        self.add_option("--log-file")
        self.add_option("--log-level")

def apply_defaults(values, opts):
    for k, v in opts.__dict__.iteritems():
        if v == None and k in values:
            opts.__dict__[k] = values[k]

# These logging functions address a logging bootstrap problem.  Before
# configuration is read, we can't configure logging.  Nonetheless,
# we'd like to log errors or warnings encountered before that point.
# These logging functions split setup into two phases, the first being
# temporary and basic, and the second being fully configured.
#
# Here's how to use them:
#
#     def main():
#         setup_initial_logging()
#
#         # load config files and command-line options
#
#         setup_operational_logging(config)
#
#         # start your business

_logging_modules = "cumin", "mint", "parsley", "rosemary", "wooly"

def setup_initial_logging():
    for name in _logging_modules:
        enable_logging(name, "warn", sys.stderr)

def setup_operational_logging(values, log_max_mb=10, log_max_archives=1):
    for name in _logging_modules:
        disable_logging(name)
        enable_logging(name, values.log_level, values.log_file, 
                             log_max_mb, log_max_archives)

    if values.debug:
        for name in _logging_modules:
            enable_logging(name, "debug", sys.stderr)
