import logging

from wooly.widgets import *
from wooly.forms import *
from wooly.resources import *
from wooly.tables import *

from cumin.objectframe import *
from cumin.objectselector import *
from cumin.stat import *
from cumin.widgets import *
from cumin.OpenFlashChart import *

strings = StringCatalog(__file__)
log = logging.getLogger("cumin.slot")

class SlotFrame(ObjectFrame):
    def __init__(self, app, name):
        cls = app.model.com_redhat_grid.Slot

        super(SlotFrame, self).__init__(app, name, cls)

class SlotDataSet(CuminSqlDataSet):
    def __init__(self, app):
        super(SlotDataSet, self).__init__(app)

        self.where_exprs.default = list()

    def sort_records(self, records):
        try:
            for rec in records:
                rec["category"], rec["category_group"] = SlotMapPage.get_cat_and_group(rec["State"], rec["Activity"])
            recs = sorted(records, key=lambda x:"%s%s%s" % 
                        (x["category_group"], x["category"], x["Name"]))
        except Exception:
            return records
        return recs

class SlotView(CuminView):
    def __init__(self, app, name, slot):
        super(SlotView, self).__init__(app, name, slot)

        self.tabs = TabbedModeSet(app, "tabs")
        self.add_child(self.tabs)

        stats = SlotStats(app, "stats", slot)
        self.tabs.add_tab(stats)

        details = CuminDetails(app, "details", slot)
        self.tabs.add_tab(details)

class SlotLoadStatSet(StatSet):
    def __init__(self, app, name, object):
        super(SlotLoadStatSet, self).__init__(app, name, object)

        self.attrs = ("CondorLoadAvg", "LoadAvg"),

class SlotStats(Widget):
    def __init__(self, app, name, slot):
        super(SlotStats, self).__init__(app, name)

        stats = SlotLoadStatSet(app, "general", slot)
        self.add_child(stats)

        job_info = SlotJobInfo(app, "job_info", slot, "JobInfo")
        self.add_child(job_info)

        chart = self.LoadChart(app, "chart", slot)
        chart.stats = ("LoadAvg", "CondorLoadAvg")
        chart.duration.param.default = "3600"
        self.add_child(chart)

    def render_title(self, session):
        return "Statistics"

    class LoadChart(StatFlashChart):
        def render_title(self, session):
            return "Load"

class SlotJobInfo(PropertySet):
    def __init__(self, app, name, slot, category):
        super(SlotJobInfo, self).__init__(app, name)

        #self.defer_enabled = True
        self.update_enabled = True
        self.category = category
        self.object = slot

    def do_get_items(self, session):
        obj = self.object.get(session)
        cls = self.app.model.get_class_by_object(obj)

        return [(x.get_title(session), x.value(session, obj), x.escape)
                for x in cls.properties if x.category == self.category]

class SlotMapPage(Page):
    """ handles the slots.vis request """

    categories = {("Owner", None): ("Unavailable", "Owner"),
                  ("Unclaimed", None): ("Available", "Unclaimed"),
                  ("Matched", None): ("Starting", "Transitioning"),
                  ("Claimed", "Idle"): ("Starting", "Transitioning"),
                  ("Claimed", "Busy"): ("Running", "Busy"),
                  ("Claimed", "Retiring"): ("Running", "Busy"),
                  ("Claimed", "Suspended"): ("Paused", "Busy"),
                  ("Preempting", None): ("Stopping", "Transitioning"),
                  ("Unknown", None): ("Unknown", "Unknown")}

    shapes = {"Unavailable": {"shape": "filled_triangle", "color": (.6,.6,.6)},
              "Available": {"shape": "box", "color": (.6,.6,.6)},
              "Stopping": {"shape": "diagonal", "color": (1,.2,.2)},
              "Starting": {"shape": "diagonal", "color": (.2,1,8)},
              "Paused": {"shape": "solid", "color": (1,.2,.2)},
              "Running": {"shape": "solid", "color": (0,.8,0)},
              "Busy": {"shape": "solid", "color": (0,.8,0)},
              "Transitioning": {"shape": "diagonal", "color": (.8,0,.8)},
              "Unclaimed": {"shape": "box", "color": (.6,.6,.6)},
              "Owner": {"shape": "filled_triangle", "color": (.6,.6,.6)},
              "Unknown": {"shape": "solid", "color": (1,0,0)}}

    interiors = {"Idle": (.7,.7,.7),
               "Busy": (0.0, 0.4, 0.0),
               "Suspended": (1,0,0),
               "Vacating": (1,.73,.367),
               "Killing": (0,0,1),
               "Benchmarking": (1,.8,.8),
               "Retiring": (.8,.2,.8),
               "Unknown": (.8,.8,.8)}

    max_width = 400
    max_png_age = 30

    # start creating artificial groupings after the number of real groups
    max_top_level_items = 500

    def __init__(self, app, name, param=None, desc=""):
        super(SlotMapPage, self).__init__(app, name)

        self.object_param = param
        self.object_description = desc

        self.zoom_level = IntegerParameter(app, "zl")
        self.zoom_level.default = 1
        self.add_parameter(self.zoom_level)

        self.category = Parameter(app, "category")
        self.add_parameter(self.category)

        self.json = Parameter(app, "json")
        self.add_parameter(self.json)

        self.sequence = IntegerParameter(app, "seq")
        self.sequence.default = 0
        self.add_parameter(self.sequence)

        self.expanded = Parameter(app, "expanded")
        self.expanded.default = ""
        self.expanded.all = "all"
        self.add_parameter(self.expanded)

        self.group = Parameter(app, "group")
        self.add_parameter(self.group)

        self.slots = SlotDataSet(app)

    @classmethod
    def get_categories(cls):
        return cls.categories

    @classmethod
    def get_unique_category_groups(cls):
        categories = cls.categories
        d = dict()
        for x in categories:
            _, group = categories[x]
            if not group == "Unknown":
                d[group] = None
        return d.keys()

    def get_content_type(self, session):
        return self.json.get(session) and "text/plain" or "image/png"

    def get_cache_control(self, session):
        return "no-cache"

    def render_category(self, session, category):
        map = HeatMapChart()
        surface = map.plot_colored_shape(self.shapes[category]["color"], 
                                         self.shapes[category]["shape"], 20, 20)
        writer = Writer()
        surface.write_to_png(writer)
        return writer.to_string()

    def do_render(self, session):
        # render the legend icon
        category = self.category.get(session)
        if category:
            return self.render_category(session, category)

        # generate the json data for the flash chart
        if self.json.get(session):
            return self.render_json(session)

        zl = self.zoom_level.get(session)

        map = HeatMapChart(self.max_width, self.max_width)

        cursor = self.slots.execute(session)
        records = cursor_to_rows(cursor)

        if len(records) == 0:
            return

        records = self.slots.sort_records(records)
        shape_info = list()
        for rec in records:
            cat = rec["category"]
            shape_info.append(self.shapes[cat])
        size = map.plot_shapes(shape_info, zl)

        args = (size, map.width, map.height, len(shape_info), map.rows, map.cols)
        cookie = "|".join((str(x) for x in args))
        session.set_cookie("slot_info", cookie)

        writer = Writer()
        map.write(writer)
        return writer.to_string()

    def split_compound_attribute(self, record, attr):
        if "|" in attr:
            attr_list = attr.split("|")
            atr_list = [record[x] for x in attr_list]
            return ":".join(atr_list)
        return ""

    def get_info_array(self, records, array, attr):
        '''
        Creates a list of record indexes for each unique value of attr
        @param records: all the records
        @param array: list of indexes to check
        @param attr: the rosemary attribute to create the array for
        '''
        d = dict()
        for i in array:
            try:
                atr = records[i][attr]
            except KeyError:
                atr = self.split_compound_attribute(records[i], attr)
            if not atr:
                atr = "Unknown %s" % attr
            if not atr in d:
                d[atr] = {"plist": list()}
                d[atr]["klist"] = [atr]
            d[atr]["plist"].append(i)
        return d

    def collapse_level(self, level_dict, count):
        collapsed = dict()

        def save_bucket(low_key, key, plist, keylist):
            if low_key is not key:
                coll_key = "%s - %s" % (low_key, key)
            else:
                coll_key = low_key
            collapsed[coll_key] = {"plist": plist,
                                   "klist": keylist}

        plist = list()
        keylist = list()
        max_edge = len(level_dict) - 1
        bucket_size = int(max_edge / count)
        edge = bucket_size
        low_key = None
        for i, key in enumerate(sorted(level_dict.keys())):
            if not low_key:
                low_key = key
            plist.extend(level_dict[key]["plist"])
            keylist.append(key)
            if i >= edge:
                save_bucket(low_key, key, plist, keylist)
                low_key = None
                plist = list()
                keylist = list()
                edge += bucket_size
                if edge > max_edge:
                    edge = max_edge

        if low_key:
            save_bucket(low_key, key, plist, keylist)
        return collapsed

    @classmethod
    def get_cat_and_group(cls, state, activity):
        act = activity
        if state in ("Owner", "Unclaimed", "Matched", "Preempting", "Unknown"):
            # activity doesn't matter for these states
            act = None
        try:
            cat, cat_group = cls.categories[(state, act)]
        except Exception, e:
            log.exception(e)
            cat, cat_group = ("%s:%s" % (state, activity), "Unknown")

        return cat, cat_group

    def get_cats(self, records, plist):
        d = dict()
        for i in plist:
            state = records[i]["State"]
            activity = records[i]["Activity"]

            cat, cat_group = SlotMapPage.get_cat_and_group(state, activity)
            if not cat_group in d:
                d[cat_group] = dict()
            if not cat in d[cat_group]:
                d[cat_group][cat] = 0
                d[cat_group]["_count"] = 0
            d[cat_group]["_count"] += 1
            d[cat_group][cat] += 1

        return d

    def render_json(self, session):
        cursor = self.slots.execute(session)
        records = cursor_to_rows(cursor)

        if len(records) == 0:
            return "[]"

        records = self.slots.sort_records(records)
        group = self.group.get(session)
        slot_count = len(records)

        root = Element()

        # Not sure these are used.  Code was modified
        # when collector filtering was removed.  If values
        # are needed here they can be synthesized in the case
        # of a collector
        root.name = self.object_description
        if self.object_param != None:
            root.value = self.object_param.get(session)._id
        else:
            root.value = ""

        root.slots = slot_count
        json = self.json.get(session)
        sequence = self.sequence.get(session)
        root.vis = json
        root.seq = sequence
        root.shape_info = self.shapes
        root.activity_colors = self.interiors
        expanded = self.expanded.get(session)
        root.tree = self.treeify(records, range(slot_count), group, expanded)
        root.back = False
        return "[%s]" % root.create()

    def treeify(self, records, plist, group, expanded):
        def add_branch(level_list, tree, back, group, key, slots):
            if len(tree):
                el = Element()
                el.tree = tree
                el.back = back
                el.name = group
                el.value = key
                el.slots = slots
                el.level = 0
                level_list.append(el)

        def gen_leaves(plist):
            level_list = list()
            for i in sorted(plist, key=lambda x:"%s%s%s" % 
                            (records[x]["category_group"], records[x]["category"], records[x]["Name"])):
                el = Element()
                el.job_id = records[i]["JobId"] and records[i]["JobId"] or ""
                el.activity = records[i]["Activity"] and records[i]["Activity"] or "Unknown"
                el.state = records[i]["State"] and records[i]["State"] or "Unknown"
                el.category = records[i]["category"]
                el.category_group = records[i]["category_group"]
                el.value = records[i]["Name"] and records[i]["Name"] or ""
                el.load_avg = records[i]["LoadAvg"] and round(records[i]["LoadAvg"], 2) or 0
                el.name = "slot"
                el.slot_id = records[i]["_id"]
                level_list.append(el)
            return level_list

        def gen_summary(plist):
            # display summary info for all the slots under this grouping
            level_list = list()
            category_groups = self.get_cats(records, plist)
            for category_group in sorted(category_groups):
                el = Element()
                el.name = "slot_info"
                el.category = category_group
                el.category_group = category_group
                el.count = category_groups[category_group]["_count"]
                try:
                    el.value = "%s.%s" % (records[plist[0]][group], category_group) # uid
                except KeyError:
                    atr = self.split_compound_attribute(records[plist[0]], group)
                    el.value = "%s.%s" % (atr, category_group)
                level_list.append(el)
            return level_list

        level_list = list()
        if not group:
            tree = gen_leaves(plist)
            add_branch(level_list, tree, False, "", "", len(plist))
            return level_list

        original_level_dict = level_dict = self.get_info_array(records, plist, group)

        if len(original_level_dict) > self.max_top_level_items:
            # collapse groups into artificial sub-groups
            level_dict = self.collapse_level(original_level_dict, self.max_top_level_items)

        for key in sorted(level_dict):
            tree = list()
            if not expanded:
                #if len(level_dict) == 1:
                #    # there is only one grouping. no need to make them click on it
                #    tree = gen_leaves(level_dict[key]["plist"])
                #    add_branch(level_list, tree, False, group, key, len(level_dict[key]["plist"]))
                #else:
                    # display the top level of groups (artificial or natural)
                tree = gen_summary(level_dict[key]["plist"])
                add_branch(level_list, tree, False, group, key, len(level_dict[key]["plist"]))
            elif expanded == key:
                # lowest level of a natural group
                if key in original_level_dict:
                    tree = gen_leaves(level_dict[key]["plist"])
                    add_branch(level_list, tree, True, group, key, len(level_dict[key]["plist"]))
                else:
                    # we want to display the groups that went into an atrificial sub-group
                    for original_key in level_dict[key]["klist"]:
                        tree = gen_summary(original_level_dict[original_key]["plist"])
                        add_branch(level_list, tree, True, group, original_key, len(original_level_dict[original_key]["plist"]))
            elif expanded in level_dict[key]["klist"]:
                # lowest level of an artificial group
                tree = gen_leaves(original_level_dict[expanded]["plist"])
                add_branch(level_list, tree, True, group, expanded, len(original_level_dict[expanded]["plist"]))

        return level_list

class SlotMap(Widget):
    def __init__(self, app, name):
        super(SlotMap, self).__init__(app, name)

        self.slots = SlotDataSet(app)

        self.slot_info = self.SlotInfo(app, "slot_info", self.slots)
        self.add_child(self.slot_info)

        self.slot_legend = self.SlotLegend(app, "slot_legend")
        self.add_child(self.slot_legend)

        self.slot_clip_size = 400

    def render_image_href(self, session):
        raise Exception("Not implemented")

    def render_job_index_param(self, session):
        return "xargs"

    def render_fullpageable(self, session):
        return "fullpageable"

    def render_slot_url(self, session):
        slot = Identifiable("XXX")
        try:
            return self.page.main.grid.slot.get_href(session, slot)
        except:
            pass

    def render_slot_clip_size(self, session):
        return self.slot_clip_size

    class SlotLegend(ItemSet):
        def __init__(self, app, name):
            super(SlotMap.SlotLegend, self).__init__(app, name)

            categories = SlotMapPage.get_unique_category_groups()
            for cat in sorted(categories):
                acat = SlotCategories(app, cat)
                self.add_child(acat)

        def do_get_items(self, session):
            return self.children

        def render_category(self, session, cat):
            return cat.name

        def render_category_items(self, session, cat):
            return cat.render(session)

    class SlotInfo(ItemSet):
        display_names = {"category": ("", "", ""),
                         "JobId": ("Job ID", "", ""),
                         "System": ("System", "", ""), 
                         "Machine": ("Machine", "", ""),
                         "State": ("State", "", ""),
                         "Activity": ("Activity", "", ""),
                         "Name": ("Name", "", ""),
                         "_id": ("Slot_ID", "slotInfo_id", "hidden_row")}

        def __init__(self, app, name, slot_data):
            super(SlotMap.SlotInfo, self).__init__(app, name)

            self.index = IntegerParameter(app, "i")
            self.add_parameter(self.index)

            self.slot = Attribute(app, "slot")
            self.add_attribute(self.slot)

            self.slot_data = slot_data

            self.info_div_tmpl = WidgetTemplate(self, "bg_html")

        def do_render(self, session):
            index = self.index.get(session)

            if index is not None:
                #self.parent.slots.limit.set(session, 1)
                #self.parent.slots.offset.set(session, index)

                cursor = self.slot_data.execute(session)
                records = cursor_to_rows(cursor)

                if len(records):
                    records = self.slot_data.sort_records(records)

                    self.slot.set(session, records[index])

                writer = Writer()
                self.info_div_tmpl.render(writer, session)
                return writer.to_string()
            else:
                return super(SlotMap.SlotInfo, self).do_render(session)

        def do_get_items(self, session):
            slot = self.slot.get(session)

            return ((
                     x == "category" and (slot["category_group"],"","title_row") or self.display_names[x],
                     slot[x])
                    for x in slot if x in self.display_names)

        def render_slot_info_url(self, session):
            return self.page.get_update_url(session, (self,))

        def render_slot_info_index(self, session):
            return self.index.path

        def render_item_value(self, session, item):
            return item[1]

        def render_item_title(self, session, item):
            return item[0][0]

        def render_item_id(self, session, item):
            return item[0][1] and "id='%s'" % item[0][1] or ""

        def render_item_row_class(self, session, item):
            return item[0][2] and "class='%s'" % item[0][2] or ""

class SlotCategories(ItemSet):
    def do_get_items(self, session):
        cat_group = self.name
        categories = SlotMapPage.get_categories()

        d = dict()
        for x in categories:
            cat, group = categories[x]
            if group is cat_group: 
                d[cat] = None

        return sorted(d.keys())

    def render_item_size(self, session, cat):
        return 20

    def render_item_title(self, session, cat):
        return cat

    def render_item_href(self, session, cat):
        params = list()
        params.append("category=%s" % cat)

        return "poolslots.vis?" + ";".join(params)
