from cumin.grid.slot import SlotMap, SlotMapPage

from cumin.objectselector import ObjectSelector, ObjectLinkColumn, ObjectTable
from cumin.objectframe import ObjectFrame
from cumin.stat import StatSet, StatFlashChart
from cumin.formats import fmt_bytes
from cumin.model import CuminStatistic
from cumin.parameters import RosemaryObjectParameter

from wooly import Session, Widget
from wooly.util import StringCatalog

strings = StringCatalog(__file__)

class ObjectSelectorTableNoCheckboxes(ObjectSelector):
    def __init__(self, app, name, cls):
        super(ObjectSelectorTableNoCheckboxes, self).__init__(app, name, cls)

    def create_table(self, app, name, cls):
        return ObjectTable(app, name, cls)

class SystemSelector(ObjectSelectorTableNoCheckboxes):
    def __init__(self, app, name):
        cls = app.model.com_redhat_sesame.Sysimage

        super(SystemSelector, self).__init__(app, name, cls)

        frame = "main.inventory.system"
        col = ObjectLinkColumn(app, "name", cls.nodeName, cls._id, frame)
        self.add_column(col)
        self.add_search_filter(col)

        self.add_attribute_column(cls.osName)
        self.add_attribute_column(cls.machine)
        self.add_attribute_column(cls.memFree)
        self.add_attribute_column(cls.loadAverage1Min)

        self.enable_csv_export()

class SystemFrame(ObjectFrame):
    def __init__(self, app, name):
        cls = app.model.com_redhat_sesame.Sysimage

        super(SystemFrame, self).__init__(app, name, cls)

        self.icon_href = "resource?name=system-36.png"

        overview = SystemStats(app, "overview", self.object)
        self.view.add_tab(overview)

class SystemGeneralStatSet(StatSet):
    fmt_as_bytes = ("memFree", "swapFree")
    def __init__(self, app, name, object):
        super(SystemGeneralStatSet, self).__init__(app, name, object)

        self.attrs = ("memFree", "swapFree",
                       "loadAverage1Min", "loadAverage5Min",
                       "loadAverage10Min", "procTotal",
                       "procRunning")

    def render_item_value(self, session, item):
        stat, value = item
        if stat.name in self.fmt_as_bytes:
            return fmt_bytes(value*1024)
        return CuminStatistic.fmt_value(value)

class SystemStats(Widget):
    def __init__(self, app, name, system):
        super(SystemStats, self).__init__(app, name)

        self.add_child(SystemGeneralStatSet(app, "stats", system))

        chart = self.MemoryFlashChart(app, "freemem", system)
        chart.chart_type = "percent"
        chart.stats = ("memFree",)
        self.add_child(chart)

        chart = StatFlashChart(app, "loadavg", system)
        chart.stats = ("loadAverage1Min",)
        self.add_child(chart)

        slots = SystemSlotMap(app, "system_slot_map", system)
        self.add_child(slots)

        slot_vis = SystemSlotFlashVis(app, "system_slot_flash", system)
        self.add_child(slot_vis)

    def render_title(self, session):
        return "Overview"

    def render_slot_job_url(self, session):
        #job = Identifiable("XXX")
        #return self.page.main.grid.job.get_href(session, job, None)
        return session.marshal() # XXX

    class MemoryFlashChart(StatFlashChart):
        def render_title(self, session, *args):
            return "Memory Usage"

        def get_href_params(self, session):
            params = super(SystemStats.MemoryFlashChart, self).get_href_params(session)
            params.append("tp=memTotal")

            return params

class SystemSlotMap(SlotMap):
    def __init__(self, app, name, system):
        super(SystemSlotMap, self).__init__(app, name)

        self.system = system
        self.slot_clip_size = 280

    def do_process(self, session):
        super(SystemSlotMap, self).do_process(session)

        system = self.system.get(session)
        self.slots.add_where_expr(session, "\"System\" = '%s'", system.nodeName)

    def render_image_href(self, session):
        system = self.system.get(session)

        page = self.app.system_slots_page
        sess = Session(page)

        #page.sysimage.set(sess, system._id)
        page.sysimage.set(sess, system)

        return sess.marshal()

    def render_slot_url(self, session):
        return ""

class SystemSlotFlashVis(SystemSlotMap):
    def render_slots_href(self, session):
        system = self.system.get(session)

        page = self.app.system_slots_page
        sess = Session(page)

        page.sysimage.set(sess, system)
        page.json.set(sess, "slots")
        page.group.set(sess, None)

        return sess.marshal()

    def render_slot_chart_width(self, session):
        return 400

    def render_slot_chart_height(self, session):
        return 400

class SystemSlotMapPage(SlotMapPage):
    def __init__(self, app, name):
        self.cls = app.model.com_redhat_sesame.Sysimage
        self.sysimage = RosemaryObjectParameter(app, "id", self.cls)

        super(SystemSlotMapPage, self).__init__(app, name, self.sysimage, "System")

        self.add_parameter(self.sysimage)

    def do_process(self, session):
        super(SystemSlotMapPage, self).do_process(session)

        sysimage = self.sysimage.get(session)
        self.slots.add_where_expr(session, "\"System\" = '%s'", sysimage.nodeName)
