from sqloperation import *
from sqlmodel import *
from sqlquery import *
from sqltype import *
import util

log = logging.getLogger("rosemary.model")

class RosemaryModel(object):
    def __init__(self):
        self._packages = list()
        self._packages_by_name = dict()

        self.sql_model = SqlModel(self)

        self.sql_logging_enabled = False

    def load_model_dir(self, path):
        assert os.path.isdir(path)

        extensions = os.path.join(path, "rosemary.xml")

        for name in os.listdir(path):
            file_path = os.path.join(path, name)

            if not os.path.isfile(file_path):
                continue

            if file_path == extensions:
                continue

            if file_path.endswith(".xml"):
                self.load_model_file(file_path)

        if os.path.isfile(extensions):
            self.load_extensions(extensions)

    def load_model_file(self, path):
        tree = ElementTree()
        file = open(path, "r")

        try:
            tree.parse(file)
            self.load(tree.getroot())
        finally:
            file.close()

    def load_extensions(self, path):
        tree = ElementTree()
        file = open(path, "r")

        try:
            tree.parse(file)
            self.extend(tree.getroot())
        finally:
            file.close()

    def load(self, elem):
        # This is an attribute on a top level tag,
        # it should correspond to something like <schema package="blah">
        attr = elem.get("package")

        if attr:
            pkg = RosemaryPackage(self, attr)
            pkg.load(elem)

        # This on the other hand is a tag itself.  It should
        # correspond to something like
        # <model>
        #    <package ....>
        #         stuff
        #    </package>
        for child in elem.findall("package"):
            pkg = RosemaryPackage(self, child.get("name"))
            pkg.load(child)

    def extend(self, elem):
        for child in elem.findall("package"):
            pkg = self._packages_by_name[child.get("name")]
            pkg.extend(child)

    def init(self):
        for pkg in self._packages:
            pkg.init()

        for pkg in self._packages:
            pkg.init_references()

    @classmethod
    def find_youngest(cls, rosemary_cls, cursor, **criteria):
        youngest = None
        if rosemary_cls._storage != "none":
            try:
                objects = rosemary_cls.get_selection(cursor, **criteria)
                if len(objects) > 0:
                    youngest = objects[0]
                    for obj in objects:
                        if obj._qmf_update_time > youngest._qmf_update_time:
                            youngest = obj
            except:
                pass
        return youngest

class RosemaryPackage(object):
    def __init__(self, model, name):
        self._model = model
        self._name = name

        # Set in the extension xml file.  This marks a package as
        # containing persistent data which should not be deleted
        # by cumin-data on startup or in association with
        # the comings and goings of agents.  Rather, it
        # should be manipulated by specific interfaces.
        # com.redhat.cumin is an example.

        # Note, "persistent" may also be set at the class
        # level in cases where packages contain both persistent
        # and non-persistent data.
        self.persistent = False

        self._model._packages.append(self)
        self._model._packages_by_name[self._name] = self

        mangled = self._name.replace(".", "_")

        if hasattr(self._model, mangled):
            raise Exception("Collision: %s" % mangled)

        setattr(self._model, mangled, self)

        self._classes = list()
        self._classes_by_name = dict()
        self._classes_by_lowercase_name = dict()

        self.sql_schema = SqlSchema(self._model.sql_model, self._name)

    def load(self, elem):
        groups_by_name = dict()

        for child in elem.findall("group"):
            name = child.get("name")
            groups_by_name[name] = child

        for child in elem.findall("class"):
            cls = RosemaryClass(self, child.get("name"))
            cls.load(child, groups_by_name)

    def extend(self, elem):
        persistent = elem.get("persistent")
        if persistent is not None:
            self.persistent = persistent == "y"

        for child in elem.findall("class"):
            cls = self._classes_by_name[child.get("name")]
            cls.extend(child)

    def init(self):
        for cls in self._classes:
            cls.init()

    def init_references(self):
        for cls in self._classes:
            cls.init_references()

    def __repr__(self):
        args = (self.__class__.__name__, self._name)
        return "%s(%s)" % args

class RosemaryClass(object):
    def __init__(self, package, name):
        self._package = package
        self._name = name

        # If _persistent is None, check_persistent will
        # return the value from self._package
        self._persistent = None

        self._package._classes.append(self)
        self._package._classes_by_name[self._name] = self
        self._package._classes_by_lowercase_name[self._name.lower()] = self

        assert not hasattr(self._package, self._name), self._name

        setattr(self._package, self._name, self)

        self._storage = "database"

        self._title = None
        self._object_title = None

        self._attributes = list()

        self._headers = list()
        self._headers_by_name = dict()

        self._references = list()
        self._references_by_name = dict()

        self._inbound_references = list()

        self._properties = list()
        self._properties_by_name = dict()

        self._statistics = list()
        self._statistics_by_name = dict()

        self._methods = list()
        self._methods_by_name = dict()

        self._indexes = list()
        self._indexes_by_name = dict()

        self._id = RosemaryAttribute(self, "_id")

        self.add_headers()

    def add_headers(self):
        name = "_qmf_agent_id"
        self._qmf_agent_id = RosemaryHeader(self, name, sql_text)
        self._qmf_agent_id.title = "Agent ID"

        name = "_qmf_object_id"
        self._qmf_object_id = RosemaryHeader(self, name, sql_text)
        self._qmf_object_id.title = "Object ID"

        name = "_qmf_create_time"
        self._qmf_create_time = RosemaryHeader(self, name, sql_timestamp)
        self._qmf_create_time.title = "Create Time"

        name = "_qmf_update_time"
        self._qmf_update_time = RosemaryHeader(self, name, sql_timestamp)
        self._qmf_update_time.title = "Update Time"

        name = "_qmf_delete_time"
        self._qmf_delete_time = RosemaryHeader(self, name, sql_timestamp)
        self._qmf_delete_time.title = "Delete Time"

        self._qmf_delete_time.optional = True

    def load(self, elem, groups_by_name):
        log.debug("Loading %s", self)

        self._storage = elem.get("storage", "database")

        assert self._storage in ("database", "none")

        for child in elem.findall("group"):
            name = child.get("name")

            try:
                self.load_class_components(groups_by_name[name])
            except KeyError:
                log.error("Reference to group '%s' invalid", name)

                raise

        self.load_class_components(elem)

    def load_class_components(self, elem):
        for child in elem.findall("property"):
            name = child.get("name")

            if child.get("references"):
                attr = RosemaryReference(self, name)
            else:
                attr = RosemaryProperty(self, name)

            attr.load(child)

        for child in elem.findall("statistic"):
            stat = RosemaryStatistic(self, child.get("name"))
            stat.load(child)

        for child in elem.findall("method"):
            meth = RosemaryMethod(self, child.get("name"))
            meth.load(child)

    def extend(self, elem):
        log.debug("Extending %s", self)

        self._title = elem.findtext("title")

        self._object_title = elem.findtext("object/title")

        persistent = elem.get("persistent")
        if persistent is not None:
            self._persistent = persistent == "y"

        for child in elem.findall("property"):
            prop = self._properties_by_name[child.get("name")]
            prop.extend(child)

        for child in elem.findall("statistic"):
            stat = self._statistics_by_name[child.get("name")]
            stat.extend(child)

        for child in elem.findall("method"):
            meth = self._methods_by_name[child.get("name")]
            meth.extend(child)

        for child in elem.findall("index"):
            idx = RosemaryIndex(self, child.get("name"))
            idx.extend(child)

    def check_persistent(self):
        if self._persistent is None:
            return self._package.persistent
        return self._persistent

    def init(self):
        log.debug("Initializing %s", self)

        if not self._title:
            self._title = generate_title(self._name)

        self.add_sql_entities()

        self._id.sql_column = self.sql_table.key_column

        self.add_samples_sql()

        for hdr in self._headers:
            hdr.init()

        for prop in self._properties:
            prop.init()

        for stat in self._statistics:
            stat.init()

        for meth in self._methods:
            meth.init()

        for idx in self._indexes:
            idx.init()

        self.add_sql_constraints()
        self.add_sql_operations()

    def init_references(self):
        for ref in self._references:
            ref.init()

    def add_sql_entities(self):
        name = "%s_id_seq" % self._name
        self.sql_sequence = SqlSequence(self._package.sql_schema, name)

        self.sql_table = SqlTable(self._package.sql_schema, self._name)

        id_col = SqlColumn(self.sql_table, "_id", sql_int8)
        self.sql_table.key_column = id_col

        name = "%s_pk" % self._name
        SqlPrimaryKeyConstraint(self.sql_table, name, (id_col,))

    def add_sql_constraints(self):
        name = "%s_qmf_id_uq" % self._name
        cols = (self._qmf_agent_id.sql_column, self._qmf_object_id.sql_column)
        SqlUniqueConstraint(self.sql_table, name, cols)

    def add_sql_operations(self):
        self.sql_get_new_id = SqlGetNewId(self.sql_table, self.sql_sequence)
        self.sql_select_by_id = SqlSelectObject(self.sql_table)
        self.sql_select_by_qmf_id = SqlSelectObjectByQmfId(self.sql_table)

        self.sql_insert_object = SqlInsertObject(self.sql_table)
        self.sql_update_object = SqlUpdateObject(self.sql_table)
        self.sql_delete_object = SqlDeleteObject(self.sql_table)

    def add_samples_sql(self):
        name = "%s_samples" % self._name
        table = SqlTable(self._package.sql_schema, name)

        self.sql_samples_table = table

        agent_col = SqlColumn(table, "_qmf_agent_id", sql_text)
        object_col = SqlColumn(table, "_qmf_object_id", sql_text)

        name = "%s_qmf_id_idx" % table._name
        SqlIndex(self._package.sql_schema, name, (agent_col, object_col))

        SqlColumn(table, "_qmf_update_time", sql_timestamp)

        self.sql_samples_insert = SqlInsertObjectSamples(table)
        self.sql_samples_delete = SqlDeleteObjectSamples(table)

    def get_object(self, cursor, **criteria):
        sql = SqlSelect(self.sql_table)

        for name in criteria:
            # XXX need to translate ref=obj args here
            
            column = self.sql_table._columns_by_name[name]
            sql.add_filter(SqlValueFilter(column))

        sql.execute(cursor, criteria)

        record = cursor.fetchone()

        if not record:
            return

        obj = RosemaryObject(self, None)

        self._set_object_attributes(obj, self.sql_table._columns, record)

        return obj

    def get_selection(self, cursor, **criteria):
        selection = list()
        sql = SqlSelect(self.sql_table)

        for name in criteria:
            # XXX need to translate ref=obj args here
            
            column = self.sql_table._columns_by_name[name]
            sql.add_filter(SqlValueFilter(column))

        sql.execute(cursor, criteria)
        
        for record in cursor.fetchall():
            obj = RosemaryObject(self, None)

            self._set_object_attributes(obj, self.sql_table._columns, record)

            selection.append(obj)

        return selection

    def get_selection_like(self, cursor, **criteria):
        selection = list()
        sql = SqlSelect(self.sql_table)

        for name in criteria:
            column = self.sql_table._columns_by_name[name]
            sql.add_filter(SqlLikeFilter(column))

        sql.execute(cursor, criteria)

        for record in cursor.fetchall():
            obj = RosemaryObject(self, None)

            self.set_object_attributes(obj, self.sql_table._columns, record)

            selection.append(obj)

        return selection

    def get_object_by_id(self, cursor, id):
        assert id

        obj = RosemaryObject(self, id)

        self.load_object_by_id(cursor, obj)

        return obj

    def get_object_by_qmf_id(self, cursor, agent_id, object_id):
        obj = RosemaryObject(self, None)
        obj._qmf_agent_id = agent_id
        obj._qmf_object_id = object_id

        self.load_object_by_qmf_id(cursor, obj)

        return obj

    def get_new_id(self, cursor):
        self.sql_get_new_id.execute(cursor)

        values = cursor.fetchone()

        assert values

        return values[0]

    def create_object(self, cursor):
        id = self.get_new_id(cursor)

        return RosemaryObject(self, id)

    def load_object_by_id(self, cursor, obj):
        assert isinstance(obj, RosemaryObject)
        assert obj._id, obj

        self.sql_select_by_id.execute(cursor, obj.__dict__)

        values = cursor.fetchone()

        if not values:
            raise RosemaryNotFound()

        self._set_object_attributes(obj, self.sql_table._columns, values)

    def load_object_by_qmf_id(self, cursor, obj):
        assert isinstance(obj, RosemaryObject)
        assert obj._qmf_agent_id, obj
        assert obj._qmf_object_id, obj

        self.sql_select_by_qmf_id.execute(cursor, obj.__dict__)

        values = cursor.fetchone()

        if not values:
            raise RosemaryNotFound()

        self._set_object_attributes(obj, self.sql_table._columns, values)

    def _set_object_attributes(self, obj, columns, values):
        assert isinstance(obj, RosemaryObject)

        for column, value in zip(columns, values):
            setattr(obj, column.name, value)

        obj._load_time = datetime.now()

    def save_object(self, cursor, obj, columns=None):
        assert isinstance(obj, RosemaryObject)
        assert obj._id, obj

        if obj._load_time or obj._save_time:
            sql = self.sql_update_object
        else:
            sql = self.sql_insert_object

        if columns is None:
            columns = self.sql_table._columns

        sql.execute(cursor, obj.__dict__, columns=columns)

        obj._save_time = datetime.now()

    def add_object_sample(self, cursor, obj, columns=None):
        assert isinstance(obj, RosemaryObject)
        assert obj._id, obj

        if columns is None:
            columns = self.sql_table._columns

        self.sql_samples_insert.execute(cursor, obj.__dict__, columns=columns)

    def delete_object(self, cursor, obj):
        assert isinstance(obj, RosemaryObject)
        assert obj._id

        self.sql_delete_object.execute(cursor, obj.__dict__)

    def delete_selection(self, cursor, **criteria):
        sql = SqlDelete(self.sql_table)

        for name in criteria:
            column = self.sql_table._columns_by_name[name]
            sql.add_filter(SqlValueFilter(column))

        sql.execute(cursor, criteria)

        return cursor.rowcount

    def __repr__(self):
        args = (self.__class__.__name__, self._package._name, self._name)
        return "%s(%s,%s)" % args

class RosemaryAttribute(object):
    def __init__(self, cls, name):
        self.cls = cls
        self.name = name

        self.cls._attributes.append(self)

        assert not hasattr(self.cls, self.name), self.name

        setattr(self.cls, self.name, self)

        self.type = None
        self.references = None
        self.access = None
        self.unit = None
        self.index = None
        self.optional = None
        self.unique = None
        self.formatter = None

        self.title = None
        self.description = None
        self.short_title = None

        self.sql_column = None

    def load(self, elem):
        self.type = elem.get("type")
        self.references = elem.get("references")
        self.access = elem.get("access")
        self.unit = elem.get("unit")
        self.index = elem.get("index", "n") == "y" and True
        self.optional = elem.get("optional", "n") == "y" and True
        self.description = clean_description(elem.get("desc"))

    def extend(self, elem):
        self.title = elem.findtext("title")
        self.short_title = elem.findtext("short")
        self.unique = elem.get("unique", "n") == "y" and True
        formatter = elem.findtext("formatter")
        if formatter:
            meth = getattr(util, formatter, None)

            if meth and callable(meth):
                self.formatter = meth

    def init(self):
        if not self.title:
            self.title = generate_title(self.name)

    def __repr__(self):
        args = (self.__class__.__name__, self.cls._name, self.name)
        return "%s(%s,%s)" % args

class RosemaryHeader(RosemaryAttribute):
    def __init__(self, cls, name, type):
        super(RosemaryHeader, self).__init__(cls, name)

        self.type = type

        self.cls._headers.append(self)
        self.cls._headers_by_name[self.name] = self

    def init(self):
        super(RosemaryHeader, self).init()

        assert not self.references

        self.sql_column = SqlColumn(self.cls.sql_table, self.name, self.type)
        self.sql_column.nullable = self.optional

class RosemaryReference(RosemaryAttribute):
    def __init__(self, cls, name):
        super(RosemaryReference, self).__init__(cls, name)

        self.cls._references.append(self)
        self.cls._references_by_name[self.name] = self

        self.that_cls = None

    def init(self):
        super(RosemaryReference, self).init()

        assert self.references

        tokens = self.references.split(":", 1)

        if len(tokens) == 2:
            pkg = self.cls._package._model._packages_by_name[tokens[0]]
            cls = tokens[1]
        else:
            pkg = self.cls._package
            cls = tokens[0]

        try:
            self.that_cls = pkg._classes_by_name[cls]
            self.that_cls._inbound_references.append(self)
        except KeyError:
            log.error("Reference to '%s' invalid", self.references)

            raise

        name = "_%s_id" % self.name
        self.name = name

        col = SqlColumn(self.cls.sql_table, name, sql_int8)
        col.foreign_key_column = self.that_cls.sql_table.key_column
        col.nullable = True

        self.sql_column = col

        name = "%s_fk" % col.name

        SqlForeignKeyConstraint(col.table, name, col, col.foreign_key_column)

class RosemaryProperty(RosemaryAttribute):
    def __init__(self, cls, name):
        super(RosemaryProperty, self).__init__(cls, name)

        self.cls._properties.append(self)
        self.cls._properties_by_name[self.name] = self

    def init(self):
        super(RosemaryProperty, self).init()

        assert not self.references

        type = sql_types_by_qmf_type_string[self.type]

        self.sql_column = SqlColumn(self.cls.sql_table, self.name, type)
        self.sql_column.nullable = self.optional

        if self.unique:
            name = "%s_%s_uq" % (self.cls._name, self.name)
            SqlUniqueConstraint(self.cls.sql_table, name, (self.sql_column,))

class RosemaryStatistic(RosemaryAttribute):
    def __init__(self, cls, name):
        super(RosemaryStatistic, self).__init__(cls, name)

        self.cls._statistics.append(self)
        self.cls._statistics_by_name[self.name] = self

    def init(self):
        super(RosemaryStatistic, self).init()

        assert not self.references

        type = sql_types_by_qmf_type_string[self.type]

        self.sql_column = SqlColumn(self.cls.sql_table, self.name, type)
        self.sql_column.nullable = True

        col = SqlColumn(self.cls.sql_samples_table, self.name, type)
        col.nullable = self.optional

class RosemaryMethod(object):
    def __init__(self, cls, name):
        self.cls = cls
        self.name = name

        self.cls._methods.append(self)
        self.cls._methods_by_name[self.name] = self

        self.description = None

        self.arguments = list()
        self.arguments_by_name = dict()

    def load(self, elem):
        self.description = clean_description(elem.get("desc"))

        for child in elem.findall("arg"):
            arg = RosemaryArgument(self, child.get("name"))
            arg.load(child)

    def extend(self, elem):
        pass

    def init(self):
        for arg in self.arguments:
            arg.init()

class RosemaryArgument(object):
    def __init__(self, meth, name):
        self.meth = meth

        self.name = name
        self.type = None
        self.direction = None
        self.description = None

        self.meth.arguments.append(self)
        self.meth.arguments_by_name[self.name] = self

    def load(self, elem):
        self.type = elem.get("type")
        self.direction = elem.get("dir")
        self.description = clean_description(elem.get("desc"))

    def extend(self, elem):
        pass

    def init(self):
        pass

class RosemaryIndex(object):
    def __init__(self, cls, name):
        self.cls = cls
        self.name = name

        self.cls._indexes.append(self)
        self.cls._indexes_by_name[self.name] = self

        self.attributes = list()

        self.sql_index = None

    def extend(self, elem):
        names = elem.get("attributes")

        for name in names.split():
            attr = getattr(self.cls, name)
            self.attributes.append(attr)

    def init(self):
        log.debug("Initializing %s", self)

        schema = self.cls._package.sql_schema
        name = "%s_%s_idx" % (self.cls._name, self.name)
        columns = [x.sql_column for x in self.attributes]

        self.sql_index = SqlIndex(schema, name, columns)

    def __repr__(self):
        args = (self.__class__.__name__, self.name)
        return "%s(%s)" % args

class RosemaryObject(object):
    def __init__(self, cls, id):
        for column in cls.sql_table._columns:
            setattr(self, column.name, None)

        self._class = cls
        self._id = id
        self._load_time = None
        self._save_time = None
        self._sample_time = None

    # XXX prefix these with _
    def load(self, cursor):
        self._class.load_object_by_id(cursor, self)

    def save(self, cursor, columns=None):
        self._class.save_object(cursor, self, columns)

    def delete(self, cursor):
        self._class.delete_object(cursor, self)

    def add_sample(self, cursor, columns=None):
        self._class.add_object_sample(cursor, self, columns)

    def fake_qmf_values(self):
        self._qmf_agent_id = "__rosemary__"
        self._qmf_object_id = str(datetime.now())
        self._qmf_create_time = datetime.now()
        self._qmf_update_time = datetime.now()

    def get_title(self):
        if self._class._object_title:
            return self._class._object_title % self.__dict__

        for attr in ("name", "Name"):
            if hasattr(self, attr):
                return self.get_formatted_value(attr)

    def get_formatted_value(self, attr):
        value = self.get_value(attr)
        formatter = None
        if attr in self._class._properties_by_name:
            formatter = self._class._properties_by_name[attr].formatter
        elif attr in self._class._statistics_by_name:
            formatter = self._class._statistics_by_name[attr].formatter
        #TODO: handle formatters headers as well
        return formatter and formatter(value) or value

    def get_value(self, attr):
        try:
            value = getattr(self, attr)
        except AttributeError:
            # there is no attr, return the value of the 1st property
            value = getattr(self, self._class._properties[0].name)

        return value

    def __repr__(self):
        name = self.__class__.__name__
        args = (name, self._class, self._id, self._save_time)
        return "%s(%s,%i,%s)" % args

class RosemaryNotFound(Exception):
    pass

def generate_title(name):
    assert name

    chars = list()
    chars.append(name[0].upper())

    name = name[1:]

    prev = None
    curr = None
    next = None

    for i in range(len(name)):
        curr = name[i]

        try:
            next = name[i + 1]
        except IndexError:
            next = None

        if curr.isupper():
            if (prev and prev.islower()) or (next and next.islower()):
                chars.append(" ")
            if next and next.islower():
                curr = curr.lower()

        chars.append(curr)

        prev = curr

    return "".join(chars)

def clean_description(description):
    return description and " ".join(description.split()) or description
