from parameters import *
from resources import *
from util import *
from wooly import *

strings = StringCatalog(__file__)

class ModeSet(Widget):
    def __init__(self, app, name):
        super(ModeSet, self).__init__(app, name)

        self.modes = list()

        self.mode = self.ModeParameter(app, "m")
        self.add_parameter(self.mode)

    def add_mode(self, mode):
        self.modes.append(mode)

        super(ModeSet, self).add_child(mode)

        if not self.mode.default:
            self.mode.default = mode

    def show_child(self, session, child):
        self.mode.set(session, child)

    def do_process(self, session, *args):
        mode = self.mode.get(session)
        mode.process(session)

    def render_content(self, session, *args):
        mode = self.mode.get(session)
        return mode.render(session)

    class ModeParameter(Parameter):
        def do_marshal(self, mode):
            return mode.name

        def do_unmarshal(self, name):
            try:
                return self.widget.children_by_name[name]
            except KeyError:
                pass

class TabbedModeSet(ModeSet):
    def __init__(self, app, name):
        super(TabbedModeSet, self).__init__(app, name)

        self.__tabs = list()
        self.__tab_tmpl = WidgetTemplate(self, "tab_html")

    def add_tab(self, tab):
        self.__tabs.append(tab)
        self.add_mode(tab)

    def render_tabs(self, session):
        writer = Writer()

        for tab in self.__tabs:
            self.__tab_tmpl.render(writer, session, tab)

        return writer.to_string()

    def render_tab_href(self, session, tab):
        branch = session.branch()
        tab.show(branch)
        return branch.marshal()

    def render_tab_class(self, session, tab):
        mode = self.mode.get(session)
        return mode is tab and "class=\"selected\"" or ""

    def render_tab_content(self, session, tab):
        return tab.render_title(session)

class RadioModeSet(TabbedModeSet):
    """ exists just to provide different class/styles """
    pass

class WidgetSet(Widget):
    def __init__(self, app, name):
        super(WidgetSet, self).__init__(app, name)

        self.__widget_tmpl = WidgetTemplate(self, "widget_html")

    def render_widgets(self, session, *args):
        writer = Writer()

        for widget in self.children:
            self.__widget_tmpl.render(writer, session, widget)

        return writer.to_string()

    def render_widget(self, session, widget):
        return widget.render(session)

# XXX get rid of this, now that we have the slightly more general
# WidgetSet
class LinkSet(Widget):
    def __init__(self, app, name):
        super(LinkSet, self).__init__(app, name)

        self.__links = list()
        self.__link_tmpl = WidgetTemplate(self, "link_html")

    def add_link(self, link):
        self.__links.append(link)
        self.add_child(link)

    def get_links(self, session):
        return self.__links

    def render_links(self, session, *args):
        writer = Writer()

        for link in self.get_links(session):
            self.__link_tmpl.render(writer, session, link)

        return writer.to_string()

    def render_link(self, session, link):
        return link.render(session)

class Link(Widget):
    def edit_session(self, session, *args):
        pass

    def render_href(self, session, *args):
        branch = session.branch()
        self.edit_session(branch, *args)
        return branch.marshal()

    def render_title(self, session, *args):
        return ""

class Toggle(Link):
    def __init__(self, app, name):
        super(Toggle, self).__init__(app, name)

        self.toggled = BooleanParameter(app, "param")
        self.add_parameter(self.toggled)

    def get(self, session):
        return self.toggled.get(session)

    def set(self, session, toggled):
        self.toggled.set(session, toggled)

    def do_process(self, session, *args):
        if self.get(session):
            self.on_click(session, *args)

    def on_click(self, session, *args):
        pass

    def render_href(self, session, *args):
        branch = session.branch()
        self.set(branch, not self.get(session))
        return branch.marshal()

    def render_state(self, session, *args):
        return self.toggled.get(session) and "on" or "off"

class ItemSet(Widget):
    def __init__(self, app, name):
        super(ItemSet, self).__init__(app, name)

        self.items = Attribute(app, "items")
        self.add_attribute(self.items)

        self.item_tmpl = WidgetTemplate(self, "item_html")

    def get_item_count(self, session):
        return len(self.get_items(session))

    def get_items(self, session):
        items = self.items.get(session)

        if items is None:
            items = self.do_get_items(session)

            if items is None:
                items = ()

        return items

    def do_get_items(self, session):
        return ()

    def render_items(self, session):
        writer = Writer()

        for item in self.get_items(session):
            self.item_tmpl.render(writer, session, item)

        return writer.to_string()

    def render_item_content(self, session, item):
        return item

    def render_item_class(self, session, item):
        return "_"

    def render_none(self, session):
        """For producing a message when the set is empty"""

        return None

class SelectionItemSet(ItemSet):
    def __init__(self, app, name, selection):
        super(SelectionItemSet, self).__init__(app, name)

        self.selection = selection

    def init(self):
        super(SelectionItemSet, self).init()

        assert isinstance(self.selection, Parameter)

    def render_item_class(self, session, item):
        if self.selection.get(session) is item:
            return "selected"
        else:
            return "_"

    def render_item_href(self, session, item):
        branch = session.branch()

        self.selection.set(branch, item)

        return branch.marshal()

class RenderingItemSet(Widget):
    def __init__(self, app, name, item_renderer):
        super(RenderingItemSet, self).__init__(app, name)

        self.item_renderer = item_renderer

        self.items = Attribute(app, "items")
        self.add_attribute(self.items)

    def get_item_count(self, session):
        return len(self.get_items(session))

    def get_items(self, session):
        items = self.items.get(session)

        if items is None:
            items = self.do_get_items(session)

            if items is None:
                items = ()

        return items

    def do_get_items(self, session):
        pass

    def render_items(self, session):
        writer = Writer()

        for item in self.get_items(session):
            self.item_renderer.render(writer, session, item)

        return writer.to_string()

class ItemRenderer(object):
    def __init__(self, widget):
        self.widget = widget

    def render_content(self, session, item):
        return item

    def render(self, writer, session, item):
        return self.render_content(session, item)

class TemplateRenderer(ItemRenderer):
    def __init__(self, widget, template_key):
        super(TemplateRenderer, self).__init__(widget)

        text = self.widget.get_string(template_key)
        self.__tmpl = ObjectTemplate(self, text)

    def render(self, writer, session, item):
        self.__tmpl.render(writer, session, item)

class ItemTree(ItemSet):
    def get_items(self, session):
        """Get the root items"""
        pass

    def get_child_items(self, session):
        pass

    def render_child_items(self, session):
        writer = Writer()

        for child in self.get_child_items(session):
            self.item_renderer.render(writer, session, child)

        return writer.to_string()

class Paginator(RenderingItemSet):
    def __init__(self, app, name):
        renderer = self.PageLink(self, "page_html")
        super(Paginator, self).__init__(app, name, renderer)

        self.page_index = IntegerParameter(app, "page")
        self.page_index.default = 0
        self.add_parameter(self.page_index)

        self.pageset_index = IntegerParameter(app, "pageset")
        self.pageset_index.default = 0
        self.add_parameter(self.pageset_index)

        self.page_count = Attribute(app, "count")
        self.add_attribute(self.page_count)

        self.page_size = 14
        self.pageset_size = 5

    def get_bounds(self, session):
        page = self.page_index.get(session)
        return self.page_size * page, self.page_size * (page + 1)

    def get_pageset_bounds(self, session):
        pageset = self.pageset_index.get(session)
        return self.pageset_size * pageset, self.pageset_size * (pageset + 1)

    def set_count(self, session, count):
        page_index = self.page_index.get(session)
        pageset_index = self.pageset_index.get(session)
        count_per_pageset = self.page_size * self.pageset_size

        # is the count too low to be on the current pageset
        if count < count_per_pageset * pageset_index:
            self.pageset_index.set(session, 0)
            self.page_index.set(session, 0)

        # is the count too low to be on the current page
        elif count < page_index * self.page_size:
            self.pageset_index.set(session, 0)
            self.page_index.set(session, 0)

        return self.page_count.set(session, count)

    def get_count(self, session):
        return self.page_count.get(session)

    def get_page_count(self, session):
        count = self.get_count(session)
        return int(math.ceil(count / float(self.page_size)))

    def get_pageset_count(self, session):
        page_count = self.get_page_count(session)
        return int(math.ceil(page_count / float(self.pageset_size)))

    def __link(self, href, content, class_=""):
        return "<a href=\"%s\"%s>%s</a>" % \
            (href, class_ and " class=\"%s\" " % class_ or " ", content)

    def render_prev_page_link(self, session, *args):
        page = self.page_index.get(session)

        if page < 1:
            html = self.__link(session.marshal(), "&lt;", "pagenav disabled")
        else:
            page -= 1
            pageset_start = self.get_pageset_bounds(session)[0]

            branch = session.branch()
            self.page_index.set(branch, page)
            if page < pageset_start:
                nindex = self.pageset_index.get(session) - 1
                self.pageset_index.set(branch, nindex)
            html = self.__link(branch.marshal(), "&lt;", "pagenav")

        return html

    def render_next_page_link(self, session, *args):
        page = self.page_index.get(session)

        if page >= self.get_page_count(session) - 1:
            html = self.__link(session.marshal(), "&gt;", "pagenav disabled")
        else:
            page += 1
            pageset_end = self.get_pageset_bounds(session)[1]

            branch = session.branch()
            self.page_index.set(branch, page)
            if page >= pageset_end: # XXX should be >? bounds func is funny
                nindex = self.pageset_index.get(session) + 1
                self.pageset_index.set(branch, nindex)
            html = self.__link(branch.marshal(), "&gt;", "pagenav")

        return html

    def render_prev_pageset_link(self, session, *args):
        pageset = self.pageset_index.get(session)

        if pageset < 1:
            html = self.__link(session.marshal(), "&lt;&lt;", "pagenav disabled")
        else:
            page_index = self.page_index.get(session)
            branch = session.branch()
            self.pageset_index.set(branch, pageset - 1)
            self.page_index.set(branch, page_index - self.pageset_size)
            html = self.__link(branch.marshal(), "&lt;&lt;", "pagenav")

        return html

    def render_next_pageset_link(self, session, *args):
        pageset = self.pageset_index.get(session)

        if pageset >= self.get_pageset_count(session) - 1:
            html = self.__link(session.marshal(), "&gt;&gt;", "pagenav disabled")
        else:
            branch = session.branch()
            self.pageset_index.set(branch, pageset + 1)
            new_page_index = self.page_index.get(branch) + self.pageset_size
            max_page_count = self.get_page_count(branch)
            if new_page_index >= max_page_count:
                new_page_index = max_page_count - 1
            self.page_index.set(branch, new_page_index)
            html = self.__link(branch.marshal(), "&gt;&gt;", "pagenav")

        return html

    def do_get_items(self, session, *args):
        count = self.page_count.get(session)

        start, end = self.get_pageset_bounds(session)
        page_count = self.get_page_count(session)

        return range(start, min(end, page_count))

    class PageLink(TemplateRenderer):
        def render_class_attr(self, session, page):
            if self.widget.page_index.get(session) == page:
                return " class=\"selected\""

        def render_href(self, session, page):
            branch = session.branch()
            self.widget.page_index.set(branch, page)
            return branch.marshal()

        def render_content(self, session, page):
            return page + 1

class PropertySet(RenderingItemSet):
    def __init__(self, app, name, item_renderer=None):
        super(PropertySet, self).__init__(app, name, item_renderer)

        if self.item_renderer is None:
            self.item_renderer = PropertyRenderer(self, "property_html")

class PropertyRenderer(TemplateRenderer):
    def render_title(self, session, prop):
        return escape_amp(prop[0])

    def render_value(self, session, prop):
        value = prop[1]
        try:
            escapable = prop[2]
        except IndexError:
            escapable = True

        if type(value) is str and escapable:
            if len(value) > 30:
                value = value[:30] + "..."

            value = escape(value)
        elif value is None:
            value = "<em>None</em>"

        return value

class ActionSet(RenderingItemSet):
    def __init__(self, app, name, item_renderer=None):
        super(ActionSet, self).__init__(app, name, item_renderer)

        if self.item_renderer is None:
            self.item_renderer = ActionRenderer(self, "action_html")

class ActionRenderer(TemplateRenderer):
    def render_href(self, session, action):
        return action[2] and " href=\"%s\"" % action[0] or ""

    def render_content(self, session, action):
        return action[1]

    def render_tag(self, session, action):
        return action[2] and "a" or "span"
